#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import TYPE_CHECKING

import numpy as np
from nomad.metainfo import MEnum, Quantity

if TYPE_CHECKING:
    from nomad.datamodel.datamodel import EntryArchive
    from nomad.metainfo import Context, Section
    from structlog.stdlib import BoundLogger


from nomad_simulations.schema_packages.atoms_state import AtomsState, OrbitalsState
from nomad_simulations.schema_packages.physical_property import PhysicalProperty
from nomad_simulations.schema_packages.utils import get_variables
from nomad_simulations.schema_packages.variables import (
    Frequency,
    ImaginaryTime,
    KMesh,
    MatsubaraFrequency,
    Time,
    WignerSeitz,
)


class BaseGreensFunction(PhysicalProperty):
    """
    A base class used to define shared commonalities between Green's function-related properties. This is the case for `ElectronicGreensFunction`,
    `ElectronicSelfEnergy`, `HybridizationFunction` in DMFT simulations.

    These physical properties are matrices matrix represented in different spaces. These spaces are stored in
    `variables` and can be: `WignerSeitz` (real space), `KMesh`, `MatsubaraFrequency`, `Frequency`, `Time`, or `ImaginaryTime`.
    For example, G(k, ω) will have `variables = [KMesh(points), RealFrequency(points)]`.

    The `rank` is determined by the number of atoms and orbitals involved in correlations, so:
        `rank = [n_atoms, n_correlated_orbitals]`

    Further information in M. Wallerberger et al., Comput. Phys. Commun. 235, 2 (2019).
    """

    # ! we use `atoms_state_ref` and `orbitals_state_ref` to enforce order in the matrices

    n_atoms = Quantity(
        type=np.int32,
        description="""
        Number of atoms involved in the correlations effect and used for the matrix representation of the property.
        """,
    )

    atoms_state_ref = Quantity(
        type=AtomsState,
        shape=['n_atoms'],
        description="""
        Reference to the `AtomsState` section in which the Green's function properties are calculated.
        """,
    )

    n_correlated_orbitals = Quantity(
        type=np.int32,
        description="""
        Number of orbitals involved in the correlations effect and used for the matrix representation of the property.
        """,
    )

    correlated_orbitals_ref = Quantity(
        type=OrbitalsState,
        shape=['n_correlated_orbitals'],
        description="""
        Reference to the `OrbitalsState` section in which the Green's function properties are calculated.
        """,
    )

    spin_channel = Quantity(
        type=np.int32,
        description="""
        Spin channel of the corresponding electronic property. It can take values of 0 and 1.
        """,
    )

    local_model_type = Quantity(
        type=MEnum('impurity', 'lattice'),
        description="""
        Type of Green's function calculated from the mapping of the local Hubbard-Kanamori model
        into the Anderson impurity model.

        The `impurity` Green's function describe the electronic correlations for the impurity, and it
        is a local function. The `lattice` Green's function includes the coupling to the lattice
        and hence it is a non-local function. In DMFT, the `lattice` term is approximated to be the
        `impurity` one, so that these simulations are converged if both types of the local
        part of the `lattice` Green's function coincides with the `impurity` Green's function.
        """,
    )

    space_id = Quantity(
        type=MEnum(
            'r',
            'rt',
            'rw',
            'rit',
            'riw',
            'k',
            'kt',
            'kw',
            'kit',
            'kiw',
            't',
            'it',
            'w',
            'iw',
        ),
        description="""
        String used to identify the space in which the Green's function property is represented. The spaces are:

        | `space_id` | `variables` |
        | ------ | ------ |
        | 'r' | WignerSeitz |
        | 'rt' | WignerSeitz + Time |
        | 'rw' | WignerSeitz + Frequency |
        | 'rit' | WignerSeitz + ImaginaryTime |
        | 'riw' | WignerSeitz + MatsubaraFrequency |
        | 'k' | KMesh |
        | 'kt' | KMesh + Time |
        | 'kw' | KMesh + Frequency |
        | 'kit' | KMesh + ImaginaryTime |
        | 'kiw' | KMesh + MatsubaraFrequency |
        | 't' | Time |
        | 'it' | Frequency |
        | 'w' | ImaginaryTime |
        | 'iw' | MatsubaraFrequency |
        """,
    )

    def __init__(
        self, m_def: 'Section' = None, m_context: 'Context' = None, **kwargs
    ) -> None:
        super().__init__(m_def, m_context, **kwargs)
        # ! n_orbitals need to be set up during initialization of the class
        self.rank = [
            int(kwargs.get('n_atoms')),
            int(kwargs.get('n_correlated_orbitals')),
        ]

    def resolve_space_id(self) -> str:
        """
        Resolves the `space_id` based on the stored `variables` in the class.

        Returns:
            str: The resolved `space_id` of the Green's function property.
        """
        _real_space_map = {
            'r': WignerSeitz,  # ? check if this is correct
            'k': KMesh,
        }
        _time_space_map = {
            't': Time,
            'it': ImaginaryTime,
            'w': Frequency,
            'iw': MatsubaraFrequency,
        }

        def find_space_id(space_map: dict) -> str:
            """
            Finds the id string for a given map.

            Args:
                space_map (dict[str, Variables]): _description_

            Returns:
                str: _description_
            """
            for space_id, variable_cls in space_map.items():
                space_variable = get_variables(
                    variables=self.variables, variable_cls=variable_cls
                )
                if len(space_variable) > 0:
                    return space_id
            return ''

        space_id = find_space_id(_real_space_map) + find_space_id(_time_space_map)
        return space_id if space_id else None

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)

        space_id = self.resolve_space_id()
        if self.space_id is not None and self.space_id != space_id:
            logger.warning(
                f'The stored `space_id`, {self.space_id}, does not coincide with the resolved one, {space_id}. We will update it.'
            )
        self.space_id = space_id


class ElectronicGreensFunction(BaseGreensFunction):
    """
    Charge-charge correlation functions.
    """

    iri = 'http://fairmat-nfdi.eu/taxonomy/ElectronicGreensFunction'

    value = Quantity(
        type=np.complex128,
        unit='1/joule',
        description="""
        Value of the electronic Green's function matrix.
        """,
    )

    def __init__(
        self, m_def: 'Section' = None, m_context: 'Context' = None, **kwargs
    ) -> None:
        super().__init__(m_def, m_context, **kwargs)
        self.name = self.m_def.name

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


class ElectronicSelfEnergy(BaseGreensFunction):
    """
    Corrections to the energy of an electron due to its interactions with its environment.
    """

    iri = 'http://fairmat-nfdi.eu/taxonomy/ElectronicSelfEnergy'

    value = Quantity(
        type=np.complex128,
        unit='joule',
        description="""
        Value of the electronic self-energy matrix.
        """,
    )

    def __init__(
        self, m_def: 'Section' = None, m_context: 'Context' = None, **kwargs
    ) -> None:
        super().__init__(m_def, m_context, **kwargs)
        self.name = self.m_def.name

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


class HybridizationFunction(BaseGreensFunction):
    """
    Dynamical hopping of the electrons in a lattice in and out of the reservoir or bath.
    """

    iri = 'http://fairmat-nfdi.eu/taxonomy/HybridizationFunction'

    value = Quantity(
        type=np.complex128,
        unit='joule',
        description="""
        Value of the electronic hybridization function.
        """,
    )

    def __init__(
        self, m_def: 'Section' = None, m_context: 'Context' = None, **kwargs
    ) -> None:
        super().__init__(m_def, m_context, **kwargs)
        self.name = self.m_def.name

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


class QuasiparticleWeight(PhysicalProperty):
    """
    Renormalization of the electronic mass due to the interactions with the environment. Within the Fermi liquid
    theory of solids, this is calculated as:

        Z = 1 - ∂Σ/∂ω|ω=0

    where Σ is the `ElectronicSelfEnergy`. The quasiparticle weight is a measure of the strength of the
    electron-electron interactions and takes values between 0 and 1, with Z = 1 representing a non-correlated
    system, and Z = 0 the Mott state.
    """

    # ! we use `atoms_state_ref` and `orbitals_state_ref` to enforce order in the matrices

    iri = 'http://fairmat-nfdi.eu/taxonomy/HybridizationFunction'

    system_correlation_strengths = Quantity(
        type=MEnum(
            'non-correlated metal',
            'strongly-correlated metal',
            'OSMI',
            'Mott insulator',
        ),
        description="""
        String used to identify the type of system based on the strength of the electron-electron interactions.

        | `type` | Description |
        | ------ | ------ |
        | 'non-correlated metal' | All `value` are above 0.7. Renormalization effects are negligible. |
        | 'strongly-correlated metal' | All `value` are below 0.4 and above 0. Renormalization effects are important. |
        | 'OSMI' | Orbital-selective Mott insulator: some orbitals have a zero `value` while others a finite one. |
        | 'Mott insulator' | All `value` are 0.0. Mott insulator state. |
        """,
    )

    n_atoms = Quantity(
        type=np.int32,
        description="""
        Number of atoms involved in the correlations effect and used for the matrix representation of the quasiparticle weight.
        """,
    )

    atoms_state_ref = Quantity(
        type=AtomsState,
        shape=['n_atoms'],
        description="""
        Reference to the `AtomsState` section in which the Green's function properties are calculated.
        """,
    )

    n_correlated_orbitals = Quantity(
        type=np.int32,
        description="""
        Number of orbitals involved in the correlations effect and used for the matrix representation of the quasiparticle weight.
        """,
    )

    correlated_orbitals_ref = Quantity(
        type=OrbitalsState,
        shape=['n_correlated_orbitals'],
        description="""
        Reference to the `OrbitalsState` section in which the Green's function properties are calculated.
        """,
    )

    spin_channel = Quantity(
        type=np.int32,
        description="""
        Spin channel of the corresponding electronic property. It can take values of 0 and 1.
        """,
    )

    value = Quantity(
        type=np.float64,
        description="""
        Value of the quasiparticle weight matrices.
        """,
    )

    def __init__(
        self, m_def: 'Section' = None, m_context: 'Context' = None, **kwargs
    ) -> None:
        super().__init__(m_def, m_context, **kwargs)
        # ! n_orbitals need to be set up during initialization of the class
        self.rank = [
            int(kwargs.get('n_atoms')),
            int(kwargs.get('n_correlated_orbitals')),
        ]
        self.name = self.m_def.name

    def is_valid_quasiparticle_weight(self) -> bool:
        """
        Check if the quasiparticle weight values are valid, i.e., if all `value` are defined between
        0 and 1.

        Returns:
            (bool): True if the quasiparticle weight is valid, False otherwise.
        """
        if (self.value < 0.0).any() or (self.value > 1.0).any():
            return False
        return True

    def resolve_system_correlation_strengths(self) -> str:
        """
        Resolves the `system_correlation_strengths` of the quasiparticle weight based on the stored `value` values.

        Returns:
            str: The resolved `system_correlation_strengths` of the quasiparticle weight.
        """
        if np.all(self.value > 0.7):
            return 'non-correlated metal'
        elif np.all((self.value < 0.4) & (self.value > 0)):
            return 'strongly-correlated metal'
        elif np.any(self.value == 0) and np.any(self.value > 0):
            return 'OSMI'
        elif np.all(self.value < 1e-2):
            return 'Mott insulator'
        return None

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)

        if self.is_valid_quasiparticle_weight() is False:
            logger.error(
                'Invalid negative quasiparticle weights found: could not validate them.'
            )
            return

        system_correlation_strengths = self.resolve_system_correlation_strengths()
        if (
            self.system_correlation_strengths is not None
            and self.system_correlation_strengths != system_correlation_strengths
        ):
            logger.warning(
                f'The stored `system_correlation_strengths`, {self.system_correlation_strengths}, does not coincide with the resolved one, {system_correlation_strengths}. We will update it.'
            )
        self.system_correlation_strengths = system_correlation_strengths



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Optional, Union

import pytest
from nomad.datamodel import EntryArchive

from nomad_simulations.schema_packages.properties import (
    QuasiparticleWeight,
)
from nomad_simulations.schema_packages.properties.greens_function import (
    BaseGreensFunction,
)
from nomad_simulations.schema_packages.variables import (
    Frequency,
    ImaginaryTime,
    KMesh,
    MatsubaraFrequency,
    Time,
    WignerSeitz,
)

from . import logger


class TestBaseGreensFunction:
    """
    Test the `BaseGreensFunction` class defined in `properties/greens_function.py`.aa
    """

    @pytest.mark.parametrize(
        'variables, result',
        [
            ([], None),
            ([WignerSeitz()], 'r'),
            ([KMesh()], 'k'),
            ([Time()], 't'),
            ([ImaginaryTime()], 'it'),
            ([Frequency()], 'w'),
            ([MatsubaraFrequency()], 'iw'),
            ([WignerSeitz(), Time()], 'rt'),
            ([WignerSeitz(), ImaginaryTime()], 'rit'),
            ([WignerSeitz(), Frequency()], 'rw'),
            ([WignerSeitz(), MatsubaraFrequency()], 'riw'),
            ([KMesh(), Time()], 'kt'),
            ([KMesh(), ImaginaryTime()], 'kit'),
            ([KMesh(), Frequency()], 'kw'),
            ([KMesh(), MatsubaraFrequency()], 'kiw'),
        ],
    )
    def test_resolve_space_id(
        self,
        variables: list[
            Union[
                WignerSeitz, KMesh, Time, ImaginaryTime, Frequency, MatsubaraFrequency
            ]
        ],
        result: str,
    ):
        """
        Test the `resolve_space_id` method of the `BaseGreensFunction` class.
        """
        gfs = BaseGreensFunction(n_atoms=1, n_correlated_orbitals=1)
        gfs.variables = variables
        assert gfs.resolve_space_id() == result

    @pytest.mark.parametrize(
        'space_id, variables, result',
        [
            ('', [], None),  # empty `space_id`
            ('rt', [], None),  # `space_id` set by parser
            ('', [WignerSeitz()], 'r'),  # resolving `space_id`
            ('rt', [WignerSeitz()], 'r'),  # normalize overwrites `space_id`
            ('', [KMesh()], 'k'),
            ('', [Time()], 't'),
            ('', [ImaginaryTime()], 'it'),
            ('', [Frequency()], 'w'),
            ('', [MatsubaraFrequency()], 'iw'),
            ('', [WignerSeitz(), Time()], 'rt'),
            ('', [WignerSeitz(), ImaginaryTime()], 'rit'),
            ('', [WignerSeitz(), Frequency()], 'rw'),
            ('', [WignerSeitz(), MatsubaraFrequency()], 'riw'),
            ('', [KMesh(), Time()], 'kt'),
            ('', [KMesh(), ImaginaryTime()], 'kit'),
            ('', [KMesh(), Frequency()], 'kw'),
            ('', [KMesh(), MatsubaraFrequency()], 'kiw'),
        ],
    )
    def test_normalize(
        self,
        space_id: str,
        variables: list[
            Union[
                WignerSeitz, KMesh, Time, ImaginaryTime, Frequency, MatsubaraFrequency
            ]
        ],
        result: Optional[str],
    ):
        """
        Test the `normalize` method of the `BaseGreensFunction` class.
        """
        gfs = BaseGreensFunction(n_atoms=1, n_correlated_orbitals=1)
        gfs.variables = variables
        gfs.space_id = space_id if space_id else None
        gfs.normalize(archive=EntryArchive(), logger=logger)
        assert gfs.space_id == result


class TestQuasiparticleWeight:
    """
    Test the `QuasiparticleWeight` class defined in `properties/greens_function.py`.
    """

    @pytest.mark.parametrize(
        'value, result',
        [
            ([[1, 0.5, -2]], False),
            ([[1, 0.5, 8]], False),
            ([[1, 0.5, 0.8]], True),
        ],
    )
    def test_is_valid_quasiparticle_weight(self, value: list[float], result: bool):
        """
        Test the `is_valid_quasiparticle_weight` method of the `QuasiparticleWeight` class.
        """
        quasiparticle_weight = QuasiparticleWeight(n_atoms=1, n_correlated_orbitals=3)
        quasiparticle_weight.value = value
        assert quasiparticle_weight.is_valid_quasiparticle_weight() == result

    @pytest.mark.parametrize(
        'value, result',
        [
            ([[1, 0.9, 0.8]], 'non-correlated metal'),
            ([[0.2, 0.3, 0.1]], 'strongly-correlated metal'),
            ([[0, 0.3, 0.1]], 'OSMI'),
            ([[0, 0, 0]], 'Mott insulator'),
            ([[1.0, 0.8, 0.2]], None),
        ],
    )
    def test_resolve_system_correlation_strengths(
        self, value: list[float], result: Optional[str]
    ):
        """
        Test the `resolve_system_correlation_strengths` method of the `QuasiparticleWeight` class.
        """
        quasiparticle_weight = QuasiparticleWeight(n_atoms=1, n_correlated_orbitals=3)
        quasiparticle_weight.value = value
        assert quasiparticle_weight.resolve_system_correlation_strengths() == result

    @pytest.mark.parametrize(
        'value, result',
        [
            ([[1, 0.5, -2]], None),
            ([[1, 0.5, 8]], None),
            ([[1, 0.9, 0.8]], 'non-correlated metal'),
            ([[0.2, 0.3, 0.1]], 'strongly-correlated metal'),
            ([[0, 0.3, 0.1]], 'OSMI'),
            ([[0, 0, 0]], 'Mott insulator'),
            ([[1.0, 0.8, 0.2]], None),
        ],
    )
    def test_normalize(self, value: list[float], result: Optional[str]):
        """
        Test the `normalize` method of the `QuasiparticleWeight` class.
        """
        quasiparticle_weight = QuasiparticleWeight(n_atoms=1, n_correlated_orbitals=3)
        quasiparticle_weight.value = value
        quasiparticle_weight.normalize(archive=EntryArchive(), logger=logger)
        assert quasiparticle_weight.system_correlation_strengths == result



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import TYPE_CHECKING

import numpy as np
from nomad.metainfo import Context, Quantity, Section, SubSection

if TYPE_CHECKING:
    from nomad.datamodel.datamodel import EntryArchive
    from nomad.metainfo import Context, Section
    from structlog.stdlib import BoundLogger

from nomad_simulations.schema_packages.physical_property import (
    PhysicalProperty,
    PropertyContribution,
)

##################
# Abstract classes
##################


class BaseForce(PhysicalProperty):
    """
    Abstract class used to define a common `value` quantity with the appropriate units
    for different types of forces, which avoids repeating the definitions for each
    force class.
    """

    value = Quantity(
        type=np.dtype(np.float64),
        unit='newton',
        description="""
        """,
    )

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


class ForceContribution(BaseForce, PropertyContribution):
    """
    Abstract class for incorporating specific force contributions to the `TotalForce`.
    The inheritance from `PropertyContribution` allows to link this contribution to a
    specific component (of class `BaseModelMethod`) of the over `ModelMethod` using the
    `model_method_ref` quantity.

    For example, for a force field calculation, the `model_method_ref` may point to a
    particular potential type (e.g., a Lennard-Jones potential between atom types X and Y),
    while for a DFT calculation, it may point to a particular electronic interaction term
    (e.g., 'XC' for the exchange-correlation term, or 'Hartree' for the Hartree term).
    Then, the contribution will be named according to this model component and the `value`
    quantity will contain the force contribution from this component evaluated over all
    relevant atoms or electrons or as a function of them.
    """

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


###################################
# List of specific force properties
###################################


class TotalForce(BaseForce):
    """
    The total force on a system. `contributions` specify individual force
    contributions to the `TotalForce`.
    """

    contributions = SubSection(sub_section=ForceContribution.m_def, repeats=True)

    def __init__(
        self, m_def: 'Section' = None, m_context: 'Context' = None, **kwargs
    ) -> None:
        super().__init__(m_def, m_context, **kwargs)
        self.name = self.m_def.name

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import TYPE_CHECKING

import numpy as np
from nomad.metainfo import Quantity

if TYPE_CHECKING:
    from nomad.datamodel.datamodel import EntryArchive
    from nomad.metainfo import Context, Section
    from structlog.stdlib import BoundLogger

from nomad_simulations.schema_packages.physical_property import PhysicalProperty
from nomad_simulations.schema_packages.properties.energies import BaseEnergy

######################################
# fundamental thermodynamic properties
######################################


class Pressure(PhysicalProperty):
    """
    The force exerted per unit area by gas particles as they collide with the walls of
    their container.
    """

    # iri = 'http://fairmat-nfdi.eu/taxonomy/Pressure' # ! Does not yet exist in taxonomy

    value = Quantity(
        type=np.float64,
        unit='pascal',
        description="""
        """,
    )

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


class Volume(PhysicalProperty):
    """
    the amount of three-dimensional space that a substance or material occupies.
    """

    #! Above description suggested for taxonomy
    # TODO check back on definition after first taxonomy version

    iri = 'http://fairmat-nfdi.eu/taxonomy/Volume'

    value = Quantity(
        type=np.float64,
        unit='m ** 3',
        description="""
        """,
    )

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


class Temperature(PhysicalProperty):
    """
    a measure of the average kinetic energy of the particles in a system.
    """

    value = Quantity(
        type=np.float64,
        unit='kelvin',
        description="""
        """,
    )

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


class Heat(BaseEnergy):
    """
    The transfer of thermal energy **into** a system.
    """

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


class Work(BaseEnergy):
    """
    The energy transferred to a system by means of force applied over a distance.
    """

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


class InternalEnergy(BaseEnergy):
    """
    The total energy contained within a system, encompassing both kinetic and potential
    energies of the particles. The change in `InternalEnergy` for some thermodynamic
    process may be expressed as the `Heat` minus the `Work`.
    """

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


class Enthalpy(BaseEnergy):
    """
    The total heat content of a system, defined as 'InternalEnergy' + 'Pressure' * 'Volume'.
    """

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


class Entropy(PhysicalProperty):
    """
    A measure of the disorder or randomness in a system.

    From a thermodynamic perspective, `Entropy` is a measure of the system's energy
    dispersal at a specific temperature, and can be interpreted as the unavailability of
    a system's thermal energy for conversion into mechanical work. For a reversible
    process, the change in `Entropy` is given mathematically by an integral over the
    infinitesimal `Heat` (i.e., thermal energy transfered into the system) divided by the
    `Temperature`.

    From a statistical mechanics viewpoint, entropy quantifies the number of microscopic
    configurations (microstates) that correspond to a thermodynamic system's macroscopic
    state, as given by the Boltzmann equation for entropy.
    """

    value = Quantity(
        type=np.float64,
        unit='joule / kelvin',
        description="""
        """,
    )

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


class GibbsFreeEnergy(BaseEnergy):
    """
    The energy available to do work in a system at constant temperature and pressure,
    given by `Enthalpy` - `Temperature` * `Entropy`.
    """

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


class HelmholtzFreeEnergy(BaseEnergy):
    """
    The energy available to do work in a system at constant volume and temperature,
    given by `InternalEnergy` - `Temperature` * `Entropy`.
    """

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


class ChemicalPotential(BaseEnergy):
    """
    Free energy cost of adding or extracting a particle from a thermodynamic system.
    """

    # ! implement `iri` and `rank` as part of `m_def = Section()`

    iri = 'http://fairmat-nfdi.eu/taxonomy/ChemicalPotential'

    def __init__(
        self, m_def: 'Section' = None, m_context: 'Context' = None, **kwargs
    ) -> None:
        super().__init__(m_def, m_context, **kwargs)
        self.rank = []
        self.name = self.m_def.name

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


class HeatCapacity(PhysicalProperty):
    """
    Amount of heat to be supplied to a material to produce a unit change in its temperature.
    """

    value = Quantity(
        type=np.float64,
        unit='joule / kelvin',
        description="""
        """,
    )

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


################################
# other thermodynamic properties
################################


class VirialTensor(BaseEnergy):
    """
    A measure of the distribution of internal forces and the overall stress within
    a system of particles. Mathematically, the virial tensor is defined as minus the sum
    of the dot product between the position and force vectors for each particle.
    The `VirialTensor` can be related to the non-ideal pressure of the system through
    the virial theorem.
    """

    def __init__(
        self, m_def: 'Section' = None, m_context: 'Context' = None, **kwargs
    ) -> None:
        super().__init__(m_def, m_context, **kwargs)
        self.rank = [3, 3]
        self.name = self.m_def.name

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


class MassDensity(PhysicalProperty):
    """
    Mass per unit volume of a material.
    """

    value = Quantity(
        type=np.float64,
        unit='kg / m ** 3',
        description="""
        """,
    )

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


# ? fit better elsewhere
class Hessian(PhysicalProperty):
    """
    A square matrix of second-order partial derivatives of a potential energy function,
    describing the local curvature of the energy surface.
    """

    value = Quantity(
        type=np.float64,
        unit='joule / m ** 2',
        description="""
        """,
    )

    def __init__(
        self, m_def: 'Section' = None, m_context: 'Context' = None, **kwargs
    ) -> None:
        super().__init__(m_def, m_context, **kwargs)
        self.rank = [3, 3]
        self.name = self.m_def.name

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import TYPE_CHECKING

import numpy as np
from nomad.metainfo import Context, Quantity, Section, SubSection

if TYPE_CHECKING:
    from nomad.datamodel.datamodel import EntryArchive
    from nomad.metainfo import Context, Section
    from structlog.stdlib import BoundLogger

from nomad_simulations.schema_packages.physical_property import (
    PhysicalProperty,
    PropertyContribution,
)

##################
# Abstract classes
##################


class BaseEnergy(PhysicalProperty):
    """
    Abstract class used to define a common `value` quantity with the appropriate units
    for different types of energies, which avoids repeating the definitions for each
    energy class.
    """

    value = Quantity(
        type=np.float64,
        unit='joule',
        description="""
        """,
    )

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


class EnergyContribution(BaseEnergy, PropertyContribution):
    """
    Abstract class for incorporating specific energy contributions to the `TotalEnergy`.
    The inheritance from `PropertyContribution` allows to link this contribution to a
    specific component (of class `BaseModelMethod`) of the over `ModelMethod` using the
    `model_method_ref` quantity.

    For example, for a force field calculation, the `model_method_ref` may point to a
    particular potential type (e.g., a Lennard-Jones potential between atom types X and Y),
    while for a DFT calculation, it may point to a particular electronic interaction term
    (e.g., 'XC' for the exchange-correlation term, or 'Hartree' for the Hartree term).
    Then, the contribution will be named according to this model component and the `value`
    quantity will contain the energy contribution from this component evaluated over all
    relevant atoms or electrons or as a function of them.
    """

    # TODO address the dual parent normalization explicity
    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


####################################
# List of specific energy properties
####################################


class FermiLevel(BaseEnergy):
    """
    Energy required to add or extract a charge from a material at zero temperature. It can be also defined as the chemical potential at zero temperature.
    """

    # ! implement `iri` and `rank` as part of `m_def = Section()`

    iri = 'http://fairmat-nfdi.eu/taxonomy/FermiLevel'

    def __init__(
        self, m_def: 'Section' = None, m_context: 'Context' = None, **kwargs
    ) -> None:
        super().__init__(m_def, m_context, **kwargs)
        self.rank = []
        self.name = self.m_def.name

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


#! The only issue with this structure is that total energy will never be a sum of its contributions,
#! since kinetic energy lives separately, but I think maybe this is ok?
class TotalEnergy(BaseEnergy):
    """
    The total energy of a system. `contributions` specify individual energetic
    contributions to the `TotalEnergy`.
    """

    # ? add a generic contributions quantity to PhysicalProperty
    contributions = SubSection(sub_section=EnergyContribution.m_def, repeats=True)

    def __init__(
        self, m_def: 'Section' = None, m_context: 'Context' = None, **kwargs
    ) -> None:
        super().__init__(m_def, m_context, **kwargs)
        self.name = self.m_def.name

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


# ? Separate quantities for nuclear and electronic KEs?
class KineticEnergy(BaseEnergy):
    """
    Physical property section describing the kinetic energy of a (sub)system.
    """

    def __init__(
        self, m_def: 'Section' = None, m_context: 'Context' = None, **kwargs
    ) -> None:
        super().__init__(m_def, m_context, **kwargs)
        self.name = self.m_def.name

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)


class PotentialEnergy(BaseEnergy):
    """
    Physical property section describing the potential energy of a (sub)system.
    """

    def __init__(
        self, m_def: 'Section' = None, m_context: 'Context' = None, **kwargs
    ) -> None:
        super().__init__(m_def, m_context, **kwargs)
        self.name = self.m_def.name

    def normalize(self, archive: 'EntryArchive', logger: 'BoundLogger') -> None:
        super().normalize(archive, logger)



from typing import TYPE_CHECKING, Any, Optional

if TYPE_CHECKING:
    import pint

import numpy as np
import pytest
from nomad.datamodel.datamodel import EntryArchive
from nomad.units import ureg

from nomad_simulations.schema_packages.atoms_state import AtomsState
from nomad_simulations.schema_packages.basis_set import (
    APWBaseOrbital,
    APWLocalOrbital,
    APWOrbital,
    APWPlaneWaveBasisSet,
    AtomCenteredBasisSet,
    BasisSetContainer,
    MuffinTinRegion,
    PlaneWaveBasisSet,
    generate_apw,
)
from nomad_simulations.schema_packages.general import Simulation
from nomad_simulations.schema_packages.model_method import BaseModelMethod, ModelMethod
from nomad_simulations.schema_packages.model_system import AtomicCell, ModelSystem
from tests.conftest import refs_apw

from . import logger


@pytest.mark.parametrize(
    'ref_cutoff_radius, cutoff_energy',
    [
        (None, None),
        (1.823 / ureg.angstrom, 500 * ureg.eV),  # reference computed by ChatGPT 4o
    ],
)
def test_cutoff(
    ref_cutoff_radius: 'pint.Quantity', cutoff_energy: 'pint.Quantity'
) -> None:
    """Test the quantitative results when computing certain plane-wave cutoffs."""
    pw = APWPlaneWaveBasisSet(cutoff_energy=cutoff_energy)
    cutoff_radius = pw.compute_cutoff_radius(cutoff_energy)

    if cutoff_radius is None:
        assert cutoff_radius is ref_cutoff_radius
    else:
        assert np.isclose(
            cutoff_radius.to(ref_cutoff_radius.units).magnitude,
            ref_cutoff_radius.magnitude,
            atol=1e-3,
        )


@pytest.mark.parametrize(
    'mts, ref_mt_r_min',
    [
        ([], None),
        ([None], None),
        ([MuffinTinRegion(radius=1.0 * ureg.angstrom)], 1.0),
        ([MuffinTinRegion(radius=r * ureg.angstrom) for r in (1.0, 2.0, 3.0)], 1.0),
    ],
)
def test_mt_r_min(mts: list[Optional[MuffinTinRegion]], ref_mt_r_min: float) -> None:
    """
    Test the computation of the minimum muffin-tin radius.
    """
    bs = BasisSetContainer(basis_set_components=mts)
    mt_r_min = bs._find_mt_r_min()

    try:
        assert mt_r_min.to('angstrom').magnitude == ref_mt_r_min
    except AttributeError:
        assert mt_r_min is ref_mt_r_min

    bs.basis_set_components.append(APWPlaneWaveBasisSet(cutoff_energy=500 * ureg('eV')))
    bs.normalize(None, logger)

    try:
        assert (
            bs.basis_set_components[-2].mt_r_min.to('angstrom').magnitude
            == ref_mt_r_min
        )
    except (IndexError, AttributeError):
        assert ref_mt_r_min is None


@pytest.mark.parametrize(
    'ref_cutoff_fractional, cutoff_energy, mt_radius',
    [
        (None, None, None),
        (None, 500.0 * ureg.eV, None),
        (None, None, 1.0),
        (1.823, 500.0 * ureg.eV, 1.0 * ureg.angstrom),
    ],
)
def test_cutoff_failure(
    ref_cutoff_fractional: float,
    cutoff_energy: 'pint.Quantity',
    mt_radius: 'pint.Quantity',
) -> None:
    """Test modes where `cutoff_fractional` is not computed."""
    pw = APWPlaneWaveBasisSet(cutoff_energy=cutoff_energy if cutoff_energy else None)
    if mt_radius is not None:
        pw.cutoff_fractional = pw.compute_cutoff_fractional(
            pw.compute_cutoff_radius(cutoff_energy), mt_radius
        )

    if ref_cutoff_fractional is None:
        assert pw.cutoff_fractional is None
    else:
        assert np.isclose(pw.cutoff_fractional, ref_cutoff_fractional, atol=1e-3)


@pytest.mark.parametrize(
    'ref_index, species_def, cutoff',
    [
        (0, {}, None),
        (1, {}, 500.0),
        (
            2,
            {
                '/data/model_system/0/cell/0/atoms_state/0': {
                    'r': 1,
                    'l_max': 2,
                    'orb_d_o': [[0]],
                    'orb_param': [[0.0]],
                }
            },
            500.0,
        ),
    ],
)
def test_full_apw(
    ref_index: int, species_def: dict[str, dict[str, Any]], cutoff: Optional[float]
) -> None:
    """Test the composite structure of APW basis sets."""
    entry = EntryArchive(
        data=Simulation(
            model_system=[
                ModelSystem(
                    cell=[AtomicCell(atoms_state=[AtomsState(chemical_symbol='H')])]
                )
            ],
            model_method=[ModelMethod(numerical_settings=[])],
        )
    )

    numerical_settings = entry.data.model_method[0].numerical_settings
    numerical_settings.append(generate_apw(species_def, cutoff=cutoff))

    # test structure
    assert numerical_settings[0].m_to_dict() == refs_apw[ref_index]


@pytest.mark.parametrize(
    'ref_n_terms, e, d_o',
    [
        (None, None, None),  # unset
        (0, [], []),  # empty
        (None, [0.0], []),  # logically inconsistent
        (1, [0.0], [0]),  # apw
        (2, 2 * [0.0], [0, 1]),  # lapw
    ],
)
def test_apw_base_orbital(ref_n_terms: Optional[int], e: list[float], d_o: list[int]):
    orb = APWBaseOrbital(energy_parameter=e, differential_order=d_o)
    assert orb.get_n_terms() == ref_n_terms


@pytest.mark.parametrize('n_terms, ref_n_terms', [(None, 1), (1, 1), (2, None)])
def test_apw_base_orbital_normalize(
    n_terms: Optional[int], ref_n_terms: Optional[int]
) -> None:
    orb = APWBaseOrbital(
        n_terms=n_terms,
        energy_parameter=[0],
        differential_order=[1],
    )
    orb.normalize(None, logger)
    assert orb.n_terms == ref_n_terms


@pytest.mark.parametrize(
    'ref_type, do',
    [
        (None, None),
        (None, []),
        (None, [0, 0, 1]),
        ('apw', [0]),
        ('lapw', [0, 1]),
        ('slapw', [0, 2]),
    ],
)
def test_apw_orbital(ref_type: Optional[str], do: Optional[int]) -> None:
    orb = APWOrbital(differential_order=do)
    assert orb.do_to_type(orb.differential_order) == ref_type


# ? necessary
@pytest.mark.parametrize(
    'ref_n_terms, e, d_o',
    [
        (None, [0.0], []),
        (1, [0.0], [0]),
        (2, 2 * [0.0], [0, 1]),
        (3, 3 * [0.0], [0, 1, 0]),
    ],
)
def test_apw_local_orbital(
    ref_n_terms: Optional[int],
    e: list[float],
    d_o: list[int],
) -> None:
    orb = APWLocalOrbital(
        energy_parameter=e,
        differential_order=d_o,
    )
    assert orb.get_n_terms() == ref_n_terms


@pytest.mark.parametrize(
    'ref_type, ref_mt_counts, ref_l_counts, species_def, cutoff',
    [
        (
            None,
            [[0, 0, 0, 0, 0]],
            [[[0, 0, 0, 0, 0]]],
            {
                'H': {
                    'r': 1.0,
                    'l_max': 0,
                    'orb_d_o': [],
                    'orb_param': [],
                    'lo_d_o': [],
                    'lo_param': [],
                }
            },
            None,
        ),
        (
            None,
            [[1, 0, 0, 0, 0]],
            [[[1, 0, 0, 0, 0]]],
            {
                'H': {
                    'r': 1.0,
                    'l_max': 0,
                    'orb_d_o': [[0]],
                    'orb_param': [[0.0]],
                    'lo_d_o': [],
                    'lo_param': [],
                }
            },
            None,
        ),
        (
            'APW-like',
            [[0, 0, 0, 0, 1]],
            [[[0, 0, 0, 0, 1]]],
            {
                'H': {
                    'r': 1.0,
                    'l_max': 0,
                    'orb_d_o': [[]],
                    'orb_param': [[]],
                    'lo_d_o': [],
                    'lo_param': [],
                }
            },
            500.0,
        ),
        (
            'APW',
            [[1, 0, 0, 0, 0]],
            [[[1, 0, 0, 0, 0]]],
            {
                'H': {
                    'r': 1.0,
                    'l_max': 1,
                    'orb_d_o': [[0]],
                    'orb_param': [[0.0]],
                    'lo_d_o': [],
                    'lo_param': [],
                }
            },
            500.0,
        ),
        (
            'LAPW',
            [[1, 0, 0, 0, 0], [0, 1, 0, 0, 0]],
            [[[1, 0, 0, 0, 0]], [[0, 1, 0, 0, 0]]],
            {
                'H': {
                    'r': 1.0,
                    'l_max': 0,
                    'orb_d_o': [[0]],
                    'orb_param': [[0.0]],
                    'lo_d_o': [],
                    'lo_param': [],
                },
                'O': {
                    'r': 2.0,
                    'l_max': 0,
                    'orb_d_o': [[0, 1]],
                    'orb_param': [2 * [0.0]],
                    'lo_d_o': [],
                    'lo_param': [],
                },
            },
            500.0,
        ),
        (
            'SLAPW',
            [[1, 0, 0, 0, 0], [0, 1, 1, 0, 0]],
            [[[1, 0, 0, 0, 0]], [[0, 1, 1, 0, 0]]],
            {
                'H': {
                    'r': 1.0,
                    'l_max': 0,
                    'orb_d_o': [[0]],
                    'orb_param': [[0.0]],
                    'lo_d_o': [],
                    'lo_param': [],
                },
                'O': {
                    'r': 2.0,
                    'l_max': 2,
                    'orb_d_o': [[0, 1], [0, 2]],
                    'orb_param': 2 * [2 * [0.0]],
                    'lo_d_o': [],
                    'lo_param': [],
                },
            },
            500.0,
        ),
    ],
)
def test_determine_apw(
    ref_type: str,
    ref_mt_counts: list[list[int]],
    ref_l_counts: list[list[list[int]]],
    species_def: dict[str, dict[str, Any]],
    cutoff: Optional[float],
) -> None:
    """Test the L-channel APW structure."""
    ref_keys = ('apw', 'lapw', 'slapw', 'lo', 'other')
    bs = generate_apw(species_def, cutoff=cutoff)

    # test from the bottom up
    for bsc in bs.basis_set_components:
        if isinstance(bsc, MuffinTinRegion):
            l_counts = ref_l_counts.pop(0)
            for l_channel in bsc.l_channels:
                try:
                    assert l_channel._determine_apw() == dict(
                        zip(ref_keys, l_counts.pop(0))
                    )
                except IndexError:
                    pass
            try:
                assert bsc._determine_apw() == dict(zip(ref_keys, ref_mt_counts.pop(0)))
            except IndexError:
                pass
    assert bs._determine_apw() == ref_type


def test_quick_step() -> None:
    """Test the feasibility of describing a QuickStep basis set."""
    entry = EntryArchive(
        data=Simulation(
            model_method=[
                ModelMethod(
                    contributions=[
                        BaseModelMethod(name='kinetic'),
                        BaseModelMethod(name='electron-ion'),
                        BaseModelMethod(name='hartree'),
                    ],
                    numerical_settings=[],
                )
            ],
        )
    )
    numerical_settings = entry.data.model_method[0].numerical_settings
    numerical_settings.append(
        BasisSetContainer(
            # scope='density',
            basis_set_components=[
                AtomCenteredBasisSet(
                    hamiltonian_scope=[
                        entry.data.model_method[0].contributions[0],
                        entry.data.model_method[0].contributions[1],
                    ],
                ),
                PlaneWaveBasisSet(
                    cutoff_energy=500 * ureg.eV,
                    hamiltonian_scope=[entry.data.model_method[0].contributions[2]],
                ),
            ],
        )
    )

    assert numerical_settings[0].m_to_dict() == {
        'm_def': 'nomad_simulations.schema_packages.basis_set.BasisSetContainer',
        'basis_set_components': [
            {
                'm_def': 'nomad_simulations.schema_packages.basis_set.AtomCenteredBasisSet',
                'hamiltonian_scope': [
                    '/data/model_method/0/contributions/0',
                    '/data/model_method/0/contributions/1',
                ],
            },
            {
                'm_def': 'nomad_simulations.schema_packages.basis_set.PlaneWaveBasisSet',
                'hamiltonian_scope': ['/data/model_method/0/contributions/2'],
                'cutoff_energy': (500.0 * ureg.eV).to('joule').magnitude,
            },
        ],
    }
    # TODO: generate a QuickStep generator in the CP2K plugin



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .band_gap import ElectronicBandGap
from .band_structure import ElectronicBandStructure, ElectronicEigenvalues, Occupancy
from .energies import (
    EnergyContribution,
    FermiLevel,
    KineticEnergy,
    PotentialEnergy,
    TotalEnergy,
)
from .fermi_surface import FermiSurface
from .forces import BaseForce, ForceContribution, TotalForce
from .greens_function import (
    ElectronicGreensFunction,
    ElectronicSelfEnergy,
    HybridizationFunction,
    QuasiparticleWeight,
)
from .hopping_matrix import CrystalFieldSplitting, HoppingMatrix
from .permittivity import Permittivity
from .spectral_profile import (
    AbsorptionSpectrum,
    DOSProfile,
    ElectronicDensityOfStates,
    SpectralProfile,
    XASSpectrum,
)
from .thermodynamics import (
    ChemicalPotential,
    Enthalpy,
    Entropy,
    GibbsFreeEnergy,
    Heat,
    HeatCapacity,
    HelmholtzFreeEnergy,
    Hessian,
    InternalEnergy,
    MassDensity,
    Pressure,
    Temperature,
    VirialTensor,
    Volume,
    Work,
)
